/*******************************************************\
 *  Funktionen zur Nutzung des RF(M)12-Funkmoduls       *
 *  ---------------------------------------------       *
 *  Basiert auf der Library von Benedikt K.             *
 *  http://www.mikrocontroller.net/topic/67273          *
 *  ---------------------------------------------       *
 *  Angepasst, erweitert und in einer Klasse gekapselt  *
 *  von Gruppen 1 und 3, Teilprojekt 2                  *
 *  Bearbeiter: Jan Unmann                              *
\*******************************************************/

#define PUFFERGROESSE_FUNK 30

#define F_CPU 9216000UL     // 9,216 MHz Quarz
#include <util/delay.h>     // Fuer _delay_ms()
#include <string.h>

#define RF12FREQ(freq)	((freq-430.0)/0.0025)   // Umrechnung Frequenz in MHz --> Einstellwert

#define RF_PORT	PORTB       // Der verwendete Port des C wird definiert
#define RF_DDR	DDRB
#define RF_PIN	PINB

#define CS      4           // Die vier Datenleitungen werden den entsprechenden Pins (von Port B) zugeordnet
#define SDI		5
#define SDO		6
#define SCK		7


/*******************************************************\
| Klasse: Funk                                           |
| Enthaelt Funktionen um das Funkmodul RF(M)12 zu nutzen |
\*******************************************************/
class Funk : public CoachBASIS, FunkBASIS
{
  private:
      char puffer[PUFFERGROESSE_FUNK];
      int index;
      int timeout;
      unsigned char i,k;
  public:

    // Allgemeine Fuktionen der Funkmodule
    unsigned short trans(unsigned short wert);			    // transfer 1 word to/from module
    void init(void);								        // initialize module
    void setfreq(unsigned short freq);					    // set center frequency
    void setbaud(unsigned short baud);					    // set baudrate
    void setpower(unsigned char power, unsigned char mod);	// set transmission settings
    void setbandwidth(unsigned char bandwidth, unsigned char gain, unsigned char drssi);	// set receiver settings
    void ready(void);								        // wait until FIFO ready (to transmit/read data)

    // Spezifische Funktionen fuer die COACH Fahrzeuge
    void start(void);                          // initialize module with COACH settings, ready to use
    void stop(void);         
    void sendeText(char *data);             // zu sendender Text mu mit \n abschliessen!
    char* empfangeText();                   //empfangener Text mu mit \n abschliessen!
    void sendePositiveInteger(int32_t zahl);
    void rx_enable(void);                                    // enables receive (in the background). Must be called before checkfor_STOP()
};

/************************\
| Funktion: init()        |
| Initialisiert das Modul |
\************************/
void Funk::init(void)
{	DDRA |= (1<<PA4);
    PORTA|= (1<<PA4);       // Power zum FUNK-Modul einschalten
    
//Auch PA7:
	DDRA |= (1<<PA7);
    PORTA|= (1<<PA7);       // Power zum FUNK-Modul einschalten
 

    RF_DDR |= (1<<SDI)|(1<<SCK)|(1<<CS);
	RF_PORT|= (1<<CS);

	for (unsigned char i=0; i<50; i++)
		_delay_ms(10);		// wait until POR done

	trans(0xC0E0);			// Set CLK output frequency: 10MHz
	trans(0x80D7);			// Enable FIFO
	trans(0xC2AB);			// Data Filter: internal
	trans(0xCA81);			// Set FIFO mode
	trans(0xE000);			// Disable wakeuptimer
	trans(0xC800);			// Disable low duty cycle
	trans(0xC4F7);			// AFC settings: autotuning: -10kHz...+7,5kHz
}

/***************************************************************\
| Funktion: start()                                |
| Initialisiert das Modul und stellt die COACH-Funkparameter ein |
\***************************************************************/
void Funk::start(void)
{	
    init();					    // Funkmodul anschalten und grundlegende Register setzen (z.B. autotuning)
	setfreq(RF12FREQ(433.12));	// Sende/Empfangsfrequenz auf 433,12MHz einstellen (weniger Strungen als der Standard bei 433,92MHz)
	setbandwidth(4, 1, 4);		// 200kHz Bandbreite, -6dB Verstrkung, DRSSI threshold: -79dBm
	setbaud(19200);			    // 19200 baud
	setpower(0, 6);			    // 1mW Ausgangangsleistung, 120kHz Frequenzshift
}

void Funk::stop(void)
{	
}


void Funk::setbandwidth(unsigned char bandwidth, unsigned char gain, unsigned char drssi)
{
	trans(0x9400|((bandwidth&7)<<5)|((gain&3)<<3)|(drssi&7));
}

void Funk::setfreq(unsigned short freq)
{	
    if (freq<96)				// 430,2400MHz
		freq=96;
	else if (freq>3903)			// 439,7575MHz
		freq=3903;
	trans(0xA000|freq);
}

void Funk::setbaud(unsigned short baud)
{	
    if (baud<664)
		baud=664;
	if (baud<5400)					        // Baudrate= 344827,58621/(R+1)/(1+CS*7)
		trans(0xC680|((43104/baud)-1));	    // R=(344828/8)/Baud-1
	else
		trans(0xC600|((344828UL/baud)-1));	// R=344828/Baud-1
}

void Funk::setpower(unsigned char power, unsigned char mod)
{	
	trans(0x9800|(power&7)|((mod&15)<<4));
}

/************************************************\
| Funktion: trans(unsigned short)                 |
| Uebertraegt ein Datenwort (16bit) zum/vom Modul |
\************************************************/
unsigned short Funk::trans(unsigned short wert)
{	
	RF_PORT &= ~(1<<CS);
	for (i=0; i<16; i++)
	{	if (wert&32768)
			RF_PORT |=  (1<<SDI);
		else
			RF_PORT &= ~(1<<SDI);
		wert<<=1;
		if (RF_PIN&(1<<SDO))
			wert|=1;
		RF_PORT |=  (1<<SCK);
		asm("nop");
		asm("nop");
		RF_PORT &= ~(1<<SCK);
	}
	RF_PORT |= (1<<CS);
	return wert;
}

/************************\
| Funktion: ready()       |
| Wartet, bis FIFO bereit |
\************************/
void Funk::ready(void)
{	
    RF_PORT &= ~(1<<SDI);
	RF_PORT &= ~(1<<CS);
	asm("nop");
	while (!(RF_PIN&(1<<SDO))); // wait until FIFO ready
}

/**************************************************\
| Funktion: sendeText(const char *)                 |
| Sendet alle Bytes des uebergebenen Arrays/Strings |
\**************************************************/
void Funk::sendeText(char *data)
{	
    trans(0x8238);          // TX on
	ready();
	trans(0xB8AA);          // Preamble: Wechselnde 1 und 0 senden (damit der Empfaenger den Takt erkennt)
	ready();
	trans(0xB8AA);          // Preamble
	ready();
	trans(0xB8AA);          // Preamble
	ready();
	trans(0xB82D);          // Sync Highbyte
	ready();
	trans(0xB8D4);          // Sync Lowbyte
	while (*data && *data!='\n')           // Sende Nutzdaten (Bytes) bis das Ende des Strings/Arrays erreicht ist, oder ein # 
	{   
        ready();
		trans(0xB800|(*data));
        *data++;
	}
	ready();
	trans(0xB800|'\n');     // Sende ein '\n', um das Ende der Uebertragung zu signalisieren
	ready();
	trans(0xB800);			// dummy data
	ready();
	trans(0xB800);			// dummy data
	ready();
	trans(0x8208);			// TX off
}

void Funk::sendePositiveInteger(int32_t zahl)
{
    index=0;            
    do
    {
        puffer[index]=48+zahl%10;
        zahl/=10;
        index++;
    } while(zahl>0);

    trans(0x8238);          // TX on
	ready();
	trans(0xB8AA);          // Preamble: Wechselnde 1 und 0 senden (damit der Empfaenger den Takt erkennt)
	ready();
	trans(0xB8AA);          // Preamble
	ready();
	trans(0xB8AA);          // Preamble
	ready();
	trans(0xB82D);          // Sync Highbyte
	ready();
	trans(0xB8D4);          // Sync Lowbyte

    for(k=0;k<index;k++)
    {
        ready();
		trans(0xB800|puffer[index-k-1]);
    }

	ready();
	trans(0xB800|'\n');     // Sende ein '\n', um das Ende der Uebertragung zu signalisieren
	ready();
	trans(0xB800);			// dummy data
	ready();
	trans(0xB800);			// dummy data
	ready();
	trans(0x8208);			// TX off
}


/*******************************************************\
| Funktion: empfangeText(unsigned char *)                 |
| Empfaengt so lange Bytes, bis ein "\n" (newline) kommt |
\*******************************************************/
char* Funk::empfangeText()
{	trans(0x82C8);			// RX on
	trans(0xCA81);			// set FIFO mode
	trans(0xCA83);			// enable FIFO
    index=0;
    do
	{   ready();
		puffer[index] = trans(0xB000);
	}while(puffer[index]!='\n' && index++<PUFFERGROESSE_FUNK);// Empfange weitere Bytes solange kein Zeilenvorschub gesendet wurde

	trans(0x8208);			// RX off
    return puffer;
}

/******************************************************************************\
| Funktion: rx_enable()                                                         |
| Stellt das Modul auf Empfang (Nachrichten werden im FIFO zwischengespeichert) |
\******************************************************************************/
void Funk::rx_enable(void)
{	
    trans(0x82C8);			// Empfang einschalten
	trans(0xCA81);			// FIFO Modus whlen
	trans(0xCA83);			// FIFO aktivieren
}
